// Property automaton -*- c++ -*-

#ifndef PROPERTY_H_
# define PROPERTY_H_
# ifdef __GNUC__
#  pragma interface
# endif // __GNUC__

# include "PropertyState.h"
# include "BitVector.h"

/** @file Property.h
 * Automaton representing the negation of a property being verified
 */

/** Property automaton */
class Property
{
public:
  /** Logical operators */
  enum Op {
    opFalse, opTrue, opNot, opAnd, opOr,
    opFinally, opGlobally, opNext, opUntil, opRelease
  };

  /** Constructor */
  Property ();
private:
  /** Copy constructor */
  explicit Property (const class Property& old);
  /** Assignment operator */
  class Property operator= (const class Property& old);
public:
  /** Destructor */
  ~Property ();

  /** Create the property automaton from an expression
   * @param translator	command line for invoking an external translator
   * @param expr	expression to be translated to an automaton
   * @return		true if the automaton was successfully generated
   */
  bool create (const char* translator,
	       class Expression& expr);

  /** Append an operator to the formula string
   * @param op		operator to be appended
   */
  void addOp (enum Op op);

  /** Append a proposition to the formula string
   * @param expr	proposition to be appended
   */
  void addExpression (class Expression& expr);

  /** Determine the index number of the initial state */
  unsigned getInitialState () const { return myInitialState; }
  /** Determine the index number of the accepting state */
  unsigned getFinalState () const { return myFinalState; }

  /** Determine the number of states in the property automaton */
  unsigned getNumStates () const { return myNumStates; }
  /** Determine the number of acceptance sets in the property automaton */
  unsigned getNumSets () const { return myNumSets; }

  /** Access a state of the automaton
   * @param i	number of the state
   * @return	the state
   */
  const class PropertyState& operator[] (unsigned i) const {
    assert (i < myNumStates);
    return myStates[i];
  }
  /** Determine whether a state belongs to an acceptance set
   * @param i	number of the state
   * @param acc	number of the acceptance set
   */
  bool accepts (unsigned i, unsigned acc) const {
    assert (!isFinite ());
    assert (myStatesAccept && i < myNumStates);
    assert (acc < myNumSets);
    return (*myStatesAccept)[myNumSets ? i * myNumSets + acc : i];
  }
  /** Determine whether the automaton is a finite-word automaton */
  bool isFinite () const { return !myNumSets; }

private:
  /** File handle for translating the formula */
  int myFD;
  /** Number of atomic propositions */
  unsigned myNumExprs;
  /** The atomic propositions (boolean expressions, the alphabet) */
  class Expression** myExprs;
  /** Number of states */
  unsigned myNumStates;
  /** The states */
  class PropertyState* myStates;
  /** Acceptance sets the states belong to */
  class BitVector* myStatesAccept;
  /** Index number of the initial state */
  unsigned myInitialState;
  /** Index number of the accepting state (for finite-word automata) */
  unsigned myFinalState;
  /** Number of acceptance sets */
  unsigned myNumSets;
};

#endif // PROPERTY_H_
