// Reporting successor states in graph-based analysis -*- c++ -*-

/** @file GraphReporter.C
 * Interface for reporting successor states in graph-based analysis
 */

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "GraphReporter.h"
#include "Graph.h"
#include "Net.h"
#include "Transition.h"
#include "Valuation.h"
#include "Printer.h"
#include "LSTS.h"
#include "PropertyState.h"
#include "GlobalMarking.h"
#ifdef EXPR_COMPILE
# include "Compilation.h"
#endif // EXPR_COMPILE

/** Report graph size statistics
 * @param states	number of generated states
 * @param arcs		number of generated arcs
 * @param pending	number of unprocessed states
 * @param printer	the output stream
 */
static void
showProgress (card_t states, card_t arcs, card_t pending,
	      const class Printer& printer)
{
  printer.print (states);
  printer.printRaw (" states, ");
  printer.print (arcs);
  printer.printRaw (" arcs, ");
  printer.print (pending);
  printer.printRaw (" pending");
  printer.finish ();
}

/** Decode a state for interpreter-based analysis
 * @param buf		the encoded inflated state
 * @param net		the model the state belongs to
 * @return		the decoded state
 */
static class GlobalMarking*
decode (const word_t* buf,
	const class Net& net)
{
  assert (!!buf);
  class GlobalMarking* marking = new class GlobalMarking (net);
  marking->decode (*net.getInitMarking (), buf);
  return marking;
}

#ifdef EXPR_COMPILE
/** Decoded a state for compiler-based analysis
 * @param buf		the encoded deflated state
 * @param size		length of the state in bytes
 * @param compilation	interface to the compiled code
 * @return		the decoded state
 */
static void
decode (word_t* buf,
	size_t size,
	const class Compilation& compilation)
{
  assert (!!buf);
  compilation.stateDecode (buf, size);
}
#endif // EXPR_COMPILE

GraphReporter::GraphReporter (
#ifdef EXPR_COMPILE
			      const class Compilation* compilation,
#endif // EXPR_COMPILE
			      const class Net& net_,
			      const class Printer& printer_,
			      unsigned maxerrors,
			      bool compress,
			      bool local,
			      unsigned thresold,
			      class Graph& graph,
			      class SearchList& states,
			      bool addOld) :
  StateReporter (
#ifdef EXPR_COMPILE
		 compilation,
#endif // EXPR_COMPILE
		 net_, printer_, maxerrors, compress, local),
  myThresold (thresold),
  myGraph (graph), myStates (states),
  myStateNum (CARD_T_MAX), myTargetNum (CARD_T_MAX),
  myWeaklyFair (new unsigned[net.getNumMaxFair () + 1]),
  myStronglyFair (new unsigned[net.getNumMaxFair () + 1]),
  myAddOld (addOld)
{
}

GraphReporter::~GraphReporter ()
{
  delete[] myWeaklyFair;
  delete[] myStronglyFair;
}

class SearchList&
GraphReporter::getSuccessors (card_t state,
			      bool modular)
{
  assert (myStates.empty ());
  if (!myGraph.isProcessed (state)) {
    myStates.push (state);
    if (modular)
      StateReporter::analyzeModular (StateReporter::Single);
    else
      StateReporter::analyze (StateReporter::Single);
  }
  else if (card_t* data = myGraph.getSuccessors (state)) {
    for (card_t i = 1; i <= *data; i++)
      myStates.push (data[i]);
    delete[] data;
  }
  return myStates;
}

card_t*
GraphReporter::prepareFair (card_t state,
			    word_t*& buf,
			    class BitUnpacker*& u)
{
  card_t* succ = myGraph.getSuccessors (state, &buf);
  if (succ) {
    assert (*succ && buf);
#ifdef EXPR_COMPILE
    if (myCompilation)
      u = 0, myCompilation->eventInflate (buf, 0);
    else
#endif // EXPR_COMPILE
      u = new class BitUnpacker (buf);
  }
  return succ;
}

unsigned*
GraphReporter::computeFair (card_t state,
			    class BitUnpacker* u,
			    unsigned** strong)
{
#ifdef EXPR_COMPILE
  if (myCompilation) {
    assert (!u);
    if (enum Error err = myCompilation->eventDecode (!!strong)) {
      printer.printRaw ("fairness constraint at state ");
      printer.printState (state);
      printer.delimiter (':');
      printer.printRaw (Valuation::msg (err));
      printer.finish ();
      flagFatal ();
      return 0;
    }
    if (strong)
      *strong = myCompilation->getStronglyFair ();
    return myCompilation->getWeaklyFair ();
  }
  else {
#endif // EXPR_COMPILE
    const class Transition* transition;
    class Valuation valuation;
    assert (!!u);
    myGraph.getSuccessor (*u, transition, valuation);
    transition->computeWeaklyFair (valuation, myWeaklyFair);
    if (!valuation.isOK ()) {
      printer.printRaw ("weakly_fair ");
    error:
      printer.printQuoted (transition->getName ());
      printer.printRaw ("at state ");
      printer.printState (state);
      printer.delimiter (':');
      printer.linebreak ();
      valuation.display (printer);
      printer.finish ();
      flagFatal ();
      return 0;
    }
    if (strong) {
      transition->computeStronglyFair (valuation, myStronglyFair);
      if (!valuation.isOK ()) {
	printer.printRaw ("strongly_fair ");
	goto error;
      }
      *strong = myStronglyFair;
    }
    return myWeaklyFair;
#ifdef EXPR_COMPILE
  }
#endif // EXPR_COMPILE
}

const unsigned*
GraphReporter::eval (card_t state,
		     unsigned prop,
		     const class PropertyState& ps)
{
#ifdef EXPR_COMPILE
  if (myCompilation) {
    size_t size;
    word_t* buf = myGraph.fetchState (state, size, 0);
    ::decode (buf, size, *myCompilation);
    delete[] buf;
    if (enum Error err = myCompilation->propEval (prop)) {
      printer.printRaw ("property ");
      printer.print (prop);
      printer.printRaw (" evaluation error at state ");
      printer.printState (state);
      printer.printRaw (":");
      printer.delimiter (' ');
      printer.printRaw (Valuation::msg (err));
      printer.finish ();
      flagFatal ();
      return 0;
    }
    return myCompilation->getPropSucc ();
  }
  else {
#endif
    /** vector of enabled gates in the property automaton */
    unsigned* enabled;
    class GlobalMarking* m = myGraph.fetchState (state);
    class Valuation valuation;
    valuation.setGlobalMarking (m);
    if (!ps.eval (valuation, enabled)) {
      printer.printRaw ("property ");
      printer.print (prop);
      printer.printRaw (" evaluation error at state ");
      printer.printState (state);
      printer.printRaw (":");
      printer.finish ();
      valuation.display (printer);
      printer.finish ();
      delete[] enabled;
      enabled = 0;
      flagFatal ();
    }
    else
      assert (valuation.isOK ());
    delete m;
    return enabled;
#ifdef EXPR_COMPILE
  }
#endif // EXPR_COMPILE
}

bool
GraphReporter::do_addState (const void* state,
			    size_t size)
{
  class Graph::AddStatus target = myGraph.add (state, size);
  myGraph.add (myStateNum, myTargetNum = target.number);

  if (myThresold && !(myGraph.getNumArcs () % myThresold))
    ::showProgress (myGraph.getNumStates (), myGraph.getNumArcs (),
		    myStates.size (), printer);

  if (target.isNew || myAddOld)
    myStates.push (target.number);

  return target.isNew;
}

word_t*
GraphReporter::do_popState (bool tail,
			    size_t& size)
{
  if (myStates.empty ())
    return 0;
  myStateNum = myStates.pop (tail);
  assert (!myGraph.isProcessed (myStateNum));
  return myGraph.fetchState (myStateNum, size, 0);
}

class GlobalMarking*
GraphReporter::pop (bool breadth)
{
  return popState (breadth)
    ? inflate (), ::decode (mySrc, net)
    : 0;
}

#ifdef EXPR_COMPILE
bool
GraphReporter::popCompiled (bool breadth)
{
  assert (!!myCompilation);
  if (popState (breadth))
    ::decode (mySrc, mySrcSize, *myCompilation);
  return !!mySrc;
}
#endif // EXPR_COMPILE

void
GraphReporter::addEvents ()
{
#ifdef EXPR_COMPILE
  if (myCompilation) {
    size_t bytes;
    const void* buf = myCompilation->eventDeflate (bytes);
    myGraph.addEvents (myStateNum, buf, bytes);
    myCompilation->eventClear ();
  }
  else {
#endif // EXPR_COMPILE
    myEventBuf.deflate ();
    myGraph.addEvents (myStateNum,
		       myEventBuf.getBuf (), myEventBuf.getNumBytes ());
    myEventBuf.clear ();
#ifdef EXPR_COMPILE
  }
#endif // EXPR_COMPILE
}

bool
GraphReporter::report (const class Transition& transition,
		       const class Valuation& valuation,
		       const class GlobalMarking& marking,
		       bool rejected)
{
  myStateBuf.deflate ();
  bool isNew;
  if (transition.getName ()) {
    const bool hidden = transition.isHidden (valuation);
    isNew = addState (myStateBuf.getBuf (), myStateBuf.getNumBytes (), hidden);
    if (!hidden)
      net.encode (myEventBuf, transition, valuation);
    if (isNew && myGraph.getLSTS ())
      myGraph.getLSTS ()->outputState (marking, myTargetNum);
  }
  else {
    // do not add arcs for anonymous (interactively defined) transitions
    class Graph::AddStatus target =
      myGraph.add (myStateBuf.getBuf (), myStateBuf.getNumBytes ());
    isNew = target.isNew;
  }

  if (rejected) {
    if (isNew) {
      myGraph.flagErroneous (myTargetNum);
      printer.printRaw (myFatal
			? "fatally rejected state "
			: "rejected state ");
      printer.printState (myTargetNum);
    }
    else {
      printer.printRaw (myFatal
			? "fatally rejected state reachable from "
			: "rejected state reachable from ");
      printer.printState (myStateNum);
    }
    printer.finish ();
    if (error ())
      flagFatal ();
  }

  return isNew;
}

void
GraphReporter::reportError (bool deadlock)
{
  if (deadlock)
    printer.printRaw (myFatal
		      ? "fatal deadlock state "
		      : "deadlock state ");
  else {
    myGraph.flagErroneous (myStateNum);
    printer.printRaw (myFatal
		      ? "fatally rejected subnet state reachable from "
		      : "rejected subnet state reachable from ");
  }
  printer.printState (myStateNum);
  printer.finish ();
  if (error ())
    flagFatal ();
}

#ifdef EXPR_COMPILE
bool
GraphReporter::report (const void* state,
		       size_t size,
		       bool rejected,
		       bool hidden)
{
  assert (!!myCompilation);
  const bool isNew = addState (state, size, hidden);

  if (isNew && myGraph.getLSTS ())
    myGraph.getLSTS ()->outputState (*myCompilation, myTargetNum);

  if (rejected) {
    myGraph.flagErroneous (myTargetNum);
    printer.printRaw (myFatal
		      ? "fatally rejected state "
		      : "rejected state ");
    printer.printState (myTargetNum);
    printer.finish ();
    if (error ())
      flagFatal ();
  }

  return isNew;
}
#endif // EXPR_COMPILE

void
GraphReporter::reject ()
{
  printer.printState (myStateNum);
  printer.delimiter (':');
}
